"""vcard package"""
__package__ = 'vcard'
__author__ = 'Victor Engmark'
__copyright__ = 'Copyright (C) 2010-2016 Victor Engmark'
__credits__ = ['Victor Engmark', 'Zearin']
__maintainer__ = 'Victor Engmark'
__email__ = 'victor.engmark@gmail.com'
__license__ = 'GPL v3 or newer'
__url__ = 'https://github.com/l0b0/vcard'
__version__ = '0.10.1'
